% Numerical and Statistical Methods 2023
% Exercise 3, Assignment 1

clear all; close all; clc

% Zero Function
f = @(x)x*exp(x)-1;

%% Q1.2,3
% Fix point functions numbered from 1 to 3
f1 = @(x)exp(-x); %(1)
f2 = @(x)(x.^2.*exp(x)+1)./exp(x)./(1+x); %(2)
f3 = @(x)x + 1 - x.*exp(x); %(3)


% initial guess of x in the range [0,1]
x0 = 0.5;

% implement fixed point iteration algorithm (function defined in the final
% section)
[x1, xvec1] = fixed_point(x0, f1);
[x2, xvec2] = fixed_point(x0, f2);
[x3, xvec3] = fixed_point(x0, f3);
% Running the iteration on phi1 and phi2 leads to convergence whereas phi3
% diverges and gives warning : convergence not achieved

% Plot convergence history (not required)
figure
plot(1:length(xvec1), abs(xvec1 - x1), 'o-')
xlim([1, length(xvec1)])
ylabel('Absolute error', 'FontSize', 16)
xlabel('Iteration step', 'FontSize', 16)
title('\phi_1 convergence history', 'FontSize', 16)

figure
plot(1:length(xvec2), abs(xvec2 - x2), 'o-')
xlim([1, length(xvec2)])
ylabel('Absolute error', 'FontSize', 16)
xlabel('Iteration step', 'FontSize', 16)
title('\phi_2 convergence history', 'FontSize', 16)
figure
%Plot every 100 steps for phi3 
plot(1:length(xvec3(1:100:end)), abs(xvec3(1:100:end) - x2), 'o-')
xlim([1, length(xvec3(1:100:end))])
ylabel('Absolute error', 'FontSize', 16)
xlabel('Iteration step', 'FontSize', 16)
title('\phi_3 convergence history', 'FontSize', 16)

%% Q1.4 Compare with fsolve
options = optimoptions('fsolve', 'StepTolerance', 1e-8);
x_fsolve = fsolve(f, x0, options);
fprintf('Relative percentage error with respect to fsolve of 1: %g \n', 100 * abs(x1 - x_fsolve) / x_fsolve)
fprintf('Relative percentage error with respect to fsolve of 2: %g \n', 100 * abs(x2 - x_fsolve) / x_fsolve)

%% Q1.5
xstar = x_fsolve;

% Calculate p and C for the two convergent functions phi1, phi2
epsi1 = abs(xvec1 - xstar);
p1 = diff(log(epsi1));
p1 = p1(2:end) ./ p1(1:end-1);
C1 = epsi1(3:end) ./ (epsi1(2:end-1).^p1);

epsi2 = abs(xvec2 - xstar);
p2 = diff(log(epsi2));
p2 = p2(2:end) ./ p2(1:end-1);
C2 = epsi2(3:end) ./ (epsi2(2:end-1).^p2);

%Plot p and C
figure
subplot(1, 2, 1);
plot(2:length(xvec1) - 1, p1, 'ro-')
hold on
yline(1, 'k--');
grid on
ylabel('Convergence order p', 'FontSize', 16)
xlabel('Iteration step', 'FontSize', 16)
title('\phi_1', 'FontSize', 16)

subplot(1, 2, 2);
grid on
plot(2:length(xvec1) - 1, C1, 'ro-')
grid on
ylabel('Rate of convergence C', 'FontSize', 16)
xlabel('Iteration step', 'FontSize', 16)
title('\phi_1', 'FontSize', 16)
%phi1 converges quite slowly (31 iter using x0 = 0.5) and the order of convergence is around 1
%(linear) for most steps


figure
subplot(1, 2, 1);
plot(2:length(xvec2) - 1, p2, 'ro-')
hold on
yline(1, 'k--');
grid on
ylabel('Convergence order p', 'FontSize', 16)
xlabel('Iteration step', 'FontSize', 16)
title('\phi_2', 'FontSize', 16)

subplot(1, 2, 2);
grid on
plot(2:length(xvec2) - 1, C2, 'ro-')
grid on
ylabel('Rate of convergence C', 'FontSize', 16)
xlabel('Iteration step', 'FontSize', 16)
title('\phi_2', 'FontSize', 16)
% phi2 converges extremely fast (5 iter using x0 = 0.5) with superlinear orders (actually >2 and >3)

%% Function Implementation : fixed point iteration

function [x, xvec] = fixed_point(x0, fun, tol, niter)
%fixed_point: Nonlinear solver using fixed point iterations.
%   [x, xvec] = fixed_point(x0, fun, tol, niter) iteratively finds fixed points
%   of the function fun (x = fun(x)).
%   Inputs:
%       x0:     Initial guess 
%       fun:    Input function (either inline, anonymous or defined in a
%               m-file)
%       tol:    Absolute tolerance (optional)
%       niter:  Maximum number of iterations (optional)
%   Output:
%       x:      Fixed point (i.e. solution of the equation x - f(x) = 0)
%       xvec:   Iteration history


%If optional parameters are not passed to the function set default values
if ~exist('tol', 'var')
    tol = 1e-8;
end
if ~exist('niter', 'var')
    niter =  1e5;
end


x_0 = x0;
x_1  = fun(x0);
xvec = [x_0, x_1];
n = 0;

while abs(x_1 - x_0) > tol && n < niter
    n = n + 1;
    xvec(n) = x_0;
    x_0 = x_1;
    x_1 = fun(x_0);
end
xvec = [xvec, x_1];

if n >= niter
    warning(sprintf('Number of iterations exceeded %G . Convergence not achieved', n));
end


x = x_1;

end